import {genId, getByteLength} from './util.js'
import { getCtx } from './als.js'

const caches = new Map()

const voidCache = {
  add() {},
  get() {},
  has() {return false},
  del() {},
  size() {return 0},
  byteLength() {return 0}
}

const getCacheFactory = (opts) => typeof opts === 'function'
  ? opts
  : getCtx().cache || createCache

export const getCache = (opts = {}) => {
  // Custom cache impl
  if (typeof opts?.get === 'function') {
    return opts
  }

  if (opts === null || !opts.ttl) {
    return voidCache
  }

  const name = opts.name || genId()

  if (caches.has(name)) {
    return caches.get(name)
  }

  const cacheFactory = getCacheFactory(opts)
  const cache = cacheFactory(opts)
  caches.set(name, cache)

  return cache
}

const hitmap = new Map()

export const hasHit = (cache, name) => hitmap.get(cache)?.has(name)

export const withCache = (cache, name, cb) => {
  if (!hitmap.has(cache)) {
    hitmap.set(cache, new Map())
  }
  const hits = hitmap.get(cache)
  if (!hits.has(name)) {
    (() => {
      let p

      hits.set(name, async () => {
        if (p) {
          return p
        }

        p = (async () => {
          if (await cache.has(name)) {
            return cache.get(name)
          }

          const value = await cb()
          await cache.add(name, value)

          p = null
          return value
        })().finally(() => hits.delete(name))

        return p
      })
    })()
  }

  return hits.get(name)()
}

export const createCache = ({ttl, evictionTimeout = ttl, warmup, limit = Infinity}) => {
  const store = new Map()
  const timer = setInterval(() => {
    const now = Date.now()
    store.forEach(({validTill, key}) => {
      if (now > validTill) {
        store.delete(key)
      }
    })
  }, evictionTimeout)
  let totalByteLength = 0

  return {
    add(key, value, _ttl) {
      const byteLength = getByteLength(value)
      if (totalByteLength + byteLength <= limit) {
        totalByteLength += byteLength
        store.set(key, {
          key,
          value,
          validTill: Date.now() + (_ttl || ttl),
          byteLength
        })
      }
      return value
    },
    has(key) {
      return store.has(key)
    },
    get(key) {
      return store.get(key)?.value || null
    },
    del(key) {
      totalByteLength -= store.get(key)?.byteLength || 0
      store.delete(key)
    },
    size() {
      return store.size
    },
    byteLength () {
      return totalByteLength
    },
    store,
    timer,
    ttl,
    warmup,
    evictionTimeout
  }
}
